"""
Glance images
"""

images = {
    "images":
        [{
            "com.rackspace__1__options": "0",
            "container_format": "ovf",
            "min_ram": 512,
            "updated_at": "2015-03-09T20:33:01Z",
            "owner": "657197",
            "file": "/v2/images/fed66ee7-fb4b-4464-b8c6-609c1fde218f/file",
            "flavor_classes": "*,!io1,!memory1,!compute1,!onmetal",
            "vm_mode": "xen",
            "com.rackspace__1__release_id": "1007",
            "com.rackspace__1__build_core": "1",
            "id": "fed66ee7-fb4b-4464-b8c6-609c1fde218f",
            "size": 1091261957,
            "os_distro": "ubuntu",
            "com.rackspace__1__release_version": "11",
            "image_type": "base",
            "self": "/v2/images/fed66ee7-fb4b-4464-b8c6-609c1fde218f",
            "disk_format": "vhd",
            "com.rackspace__1__platform_target": "PublicCloud",
            "com.rackspace__1__build_managed": "1",
            "org.openstack__1__architecture": "x64",
            "schema": "/v2/schemas/image",
            "status": "active",
            "com.rackspace__1__visible_core": "1",
            "tags": [],
            "com.rackspace__1__release_build_date": "2015-03-05_17-43-17",
            "visibility": "public",
            "auto_disk_config": "True",
            "min_disk": 20,
            "org.openstack__1__os_distro": "com.ubuntu",
            "com.rackspace__1__visible_managed": "1",
            "com.rackspace__1__source": "kickstart",
            "name": "Ubuntu 14.04 LTS (Trusty Tahr) (PV)",
            "com.rackspace__1__build_rackconnect": "1",
            "checksum": "29f8698b79091e8dc524381e562524eb",
            "created_at": "2015-03-05T18:29:51Z",
            "cache_in_nova": "True",
            "protected": "false",
            "com.rackspace__1__visible_rackconnect": "1",
            "os_type": "linux",
            "org.openstack__1__os_version": "14.04"
        }],
    "schema": "/v2/schemas/images",
    "first": "/v2/images?limit=1000"
}


image_schema = {
    "additionalProperties": {
        "type": "string"
    },
    "name": "image",
    "links": [
        {
            "href": "{self}",
            "rel": "self"
        },
        {
            "href": "{file}",
            "rel": "enclosure"
        },
        {
            "href": "{schema}",
            "rel": "describedby"
        }
    ],
    "properties": {
        "status": {
            "enum": [
                "queued",
                "saving",
                "active",
                "killed",
                "deleted",
                "pending_delete"
            ],
            "type": "string",
            "description": "Status of the image (READ-ONLY)"
        },
        "schema": {
            "type": "string",
            "description": "(READ-ONLY)"
        },
        "file": {
            "type": "string",
            "description": "(READ-ONLY)"
        },
        "direct_url": {
            "type": "string",
            "description": "URL to access the image file kept in external store (READ-ONLY)"
        },
        "name": {
            "type": ["null",
                     "string"],
            "description": "Descriptive name for the image",
            "maxLength": 255},
        "tags": {
            "items": {
                "type": "string",
                "maxLength": 255},
            "type": "array",
            "description": "List of strings related to the image"
        },
        "locations": {
            "items": {
                "required": ["url",
                             "metadata"],
                "type": "object",
                "properties": {
                    "url": {
                        "type": "string",
                        "maxLength": 255},
                    "metadata": {
                        "type": "object"
                    }}},
            "type": "array",
            "description": "A set of URLs to access the image file kept in external store"
        },
        "checksum": {
            "type": [
                "null",
                "string"
            ],
            "description": "md5 hash of image contents. (READ-ONLY)",
            "maxLength": 32
        },
        "created_at": {
            "type": "string",
            "description": "Date and time of image registration (READ-ONLY)"
        },
        "disk_format": {
            "enum": [
                None,
                "ami",
                "ari",
                "aki",
                "vhd",
                "vmdk",
                "raw",
                "qcow2",
                "vdi",
                "iso"
            ],
            "type": [
                "null",
                "string"
            ],
            "description": "Format of the disk"
        },
        "updated_at": {
            "type": "string",
            "description": "Date and time of the last image modification (READ-ONLY)"
        },
        "visibility": {
            "enum": [
                "public",
                "private"
            ],
            "type": "string",
            "description": "Scope of image accessibility"
        },
        "self": {
            "type": "string",
            "description": "(READ-ONLY)"
        },
        "min_disk": {
            "type": "integer",
            "description": "Amount of disk space (in GB) required to boot image."
        },
        "protected": {
            "type": "boolean",
            "description": "If true, image will not be deletable."
        },
        "min_ram": {
            "type": "integer",
            "description": "Amount of ram (in MB) required to boot image."
        },
        "container_format": {
            "enum": [
                None,
                "ami",
                "ari",
                "aki",
                "bare",
                "ovf",
                "ova"
            ],
            "type": [
                "null",
                "string"
            ],
            "description": "Format of the container"
        },
        "owner": {
            "type": ["null",
                     "string"],
            "description": "Owner of the image",
            "maxLength": 255},
        "virtual_size": {
            "type": ["null",
                     "integer"],
            "description": "Virtual size of image in bytes (READ-ONLY)"
        },
        "id": {
            "pattern": "^([0-9a-fA-F]){8}-([0-9a-fA-F]){4}-([0-9a-fA-F]){4}-([0-9a-fA-F])" +
                       "{4}-([0-9a-fA-F]){12}$",
            "type": "string",
            "description": "An identifier for the image"
        },
        "size": {
            "type": [
                "null",
                "integer"
            ],
            "description": "Size of image file in bytes (READ-ONLY)"
        }}}



# -*- test-case-name: mimic.test.test_session -*-

"""
Implementation of simple in-memory session storage and generation for Mimic.
"""

from six import text_type
from uuid import uuid4
from datetime import datetime, timedelta

from characteristic import attributes, Attribute


@attributes(['username', 'token', 'tenant_id', 'expires',
             Attribute('impersonator_session_map', default_factory=dict),
             Attribute('_api_objects', default_factory=dict)])
class Session(object):
    """
    A mimic Session is a record of an authentication token for a particular
    username and tenant_id.
    """

    @property
    def user_id(self):
        """
        Return a unique numeric ID based on the username.
        """
        return text_type(hash(self.username))

    def impersonator_session_for_token(self, impersonated_token):
        """
        :param impersonated_token: impersonation token for a user.
        Returns the impersonator session for the given impersonation
        token.
        """
        return self.impersonator_session_map.get(impersonated_token)

    def data_for_api(self, api_mock, data_factory):
        """
        Get the application data for a given API, creating it if necessary.
        """
        if api_mock not in self._api_objects:
            self._api_objects[api_mock] = data_factory()
        return self._api_objects[api_mock]


@attributes([Attribute('session', instance_of=Session),
             'desired_tenant'])
class NonMatchingTenantError(Exception):
    """
    A session's tenant ID does not match the desired tenant ID.
    """


class SessionStore(object):
    """
    A collection of sessions addressable by multiple different keys.

    Unlike many traditional types of session storage, new authenticated users
    are created on demand, since all authentication succeeds by default within
    Mimic.

    :ivar IReactorTime clock: The clock used to track session expiration.
    """

    def __init__(self, clock):
        """
        Create a session store with the given IReactorTime provider.
        """
        self.clock = clock
        self._token_to_session = {
            # mapping of token (unicode) to session (Session)
        }
        self._userid_to_session = {
            # mapping of userid (ascii) to session (Session)
        }
        self._tenant_to_session = {
            # mapping of tenant_id (unicode) to session (Session)
        }
        self._username_to_token = {
            # mapping of token (unicode) to username (unicode: key in
            # _token_to_session)
        }

    def _new_session(self, username_key=None, **attributes):
        """
        Create a new session and persist it according to its username and token
        values.

        :param attributes: Keyword parameters containing zero or more of
            ``username``, ``token``, and ``tenant_id``.  Any fields that are
            not specified will be filled out automatically.

        :return: A new session with all fields filled out and an expiration
                 time 1 day in the future (according to the clock associated
                 with this :obj:`MimicCore`).
        :rtype: :obj:`Session`
        """
        for key in ['username', 'token', 'tenant_id']:
            if attributes.get(key, None) is None:
                attributes[key] = key + "_" + text_type(uuid4())
                if key == 'tenant_id':
                    # integer tenant IDs - uuid4 ints are too long
                    attributes[key] = text_type(int(uuid4().int % 1e15))

        if 'expires' not in attributes:
            attributes['expires'] = (
                datetime.utcfromtimestamp(self.clock.seconds())
                + timedelta(days=1)
            )

        session = Session(**attributes)
        if username_key is None:
            username_key = session.username
        self._username_to_token[username_key] = session.token
        self._token_to_session[session.token] = session
        self._userid_to_session[session.user_id] = session
        self._tenant_to_session[session.tenant_id] = session
        return session

    def _assert_tenant_matches(self, session, tenant_id):
        if tenant_id is not None and session.tenant_id != tenant_id:
            raise NonMatchingTenantError(session=session, desired_tenant=tenant_id)

    def session_for_token(self, token, tenant_id=None):
        """
        :param unicode token: An authentication token previously created by
            session_for_api_key or session_for_username_password.

        :return: a session for the given token, only if that token already
                 exists.
        :rtype: Session

        :raise: :obj:`KeyError` if no such thing exists.
        """
        if token in self._token_to_session:
            s = self._token_to_session[token]
            self._assert_tenant_matches(s, tenant_id)
        elif tenant_id and tenant_id in self._tenant_to_session:
            s = self._tenant_to_session[tenant_id]
        else:
            s = self._new_session(token=token, tenant_id=tenant_id)
        return s

    def session_for_api_key(self, username, api_key, tenant_id=None):
        """
        Create or return a :obj:`Session`.

        :param unicode username: A user name.
        :param unicode api_key: An API key that should match the username.

        :return: a session for the given user.
        :rtype: Session
        """
        # One day, API keys will be different from passwords, but not today.
        return self.session_for_username_password(username, api_key, tenant_id)

    def session_for_username_password(self, username, password,
                                      tenant_id=None):
        """
        Create or return a :obj:`Session` based on a user's credentials.
        """
        if username in self._username_to_token:
            s = self._token_to_session[self._username_to_token[username]]
            self._assert_tenant_matches(s, tenant_id)
            return s

        if tenant_id and tenant_id in self._tenant_to_session:
            return self._tenant_to_session[tenant_id]

        return self._new_session(username=username,
                                 tenant_id=tenant_id)

    def session_for_impersonation(self, username, expires_in, impersonator_token=None,
                                  impersonated_token=None):
        """
        Create or return a :obj:`Session` impersonating a given user; this
        session updates the expiration to be that indicated.
        """
        impersonator_session = self._token_to_session.get(impersonator_token)
        session = self.session_for_username_password(
            username, "lucky we don't check passwords, isn't it"
        )
        session.expires = datetime.utcfromtimestamp(self.clock.seconds() + expires_in)
        session.impersonator_session_map[impersonated_token] = impersonator_session
        self._token_to_session[impersonated_token] = session
        return session

    def session_for_tenant_id(self, tenant_id, token_id=None):
        """
        Looks up a session based on the tenant_id.
        :param unicode tenant_id: The tenant_id of a previously-created
            session.
        :param unicode token_id: Sets token in the session to the token_id provided,
            else, creates one.
        """
        if tenant_id not in self._tenant_to_session:
            return self._new_session(tenant_id=tenant_id, token=token_id)
        return self._tenant_to_session[tenant_id]



# -*- test-case-name: mimic.test.test_noit -*-
"""
Defines get token, impersonation
"""
import xmltodict
from uuid import UUID
from mimic.rest.mimicapp import MimicApp
from mimic.canned_responses.noit import (create_check, get_check,
                                         get_all_checks, delete_check,
                                         test_check)


# TO DO:
# https://github.com/rackerlabs/mimic/issues/203


class NoitApi(object):

    """
    Rest endpoints for mocked Noit api.
    """

    app = MimicApp()

    def __init__(self, core, clock):
        """
        :param MimicCore core: The core to which this NoitApi will be
            communicating.
        """
        self.core = core
        self.clock = clock

    def validate_check_payload(self, request):
        """
        Validate the check request payload and returns the response code
        """
        content = str(request.content.read())
        try:
            payload = xmltodict.parse(content)
        except:
            return 500, None
        attributes = ["name", "module", "target", "period", "timeout",
                      "filterset"]
        for each in attributes:
            if not payload["check"]["attributes"].get(each):
                return 404, None
        return 200, payload["check"]["attributes"]

    @app.route('/checks/test', methods=['POST'])
    def test_check(self, request):
        """
        Validates the check xml and returns the metrics
        """
        response = self.validate_check_payload(request)
        if (response[0] == 200):
            request.setHeader("content-type", "application/xml")
            response_body = test_check(response[1]["module"])
            return xmltodict.unparse(response_body)
        request.setResponseCode(response[0])
        return

    @app.route('/checks/set/<check_id>', methods=['PUT'])
    def set_check(self, request, check_id):
        """
        Creates a check for the given check_id. If the check_id already
        exists, then it updates that check.
        TBD: Include error 400 and 500s. Module cannot be updated (test
            against noit service to see the response code expected)
        """
        try:
            UUID(check_id)
        except (ValueError, AttributeError):
            request.setResponseCode(500)
            return
        request.setHeader("content-type", "application/xml")
        response = self.validate_check_payload(request)
        request.setResponseCode(response[0])
        if (response[0] == 200):
            response_body = create_check(response[1], check_id)
            return xmltodict.unparse(response_body)
        return

    @app.route('/checks/show/<check_id>', methods=['GET'])
    def get_checks(self, request, check_id):
        """
        Return the current configuration and state of the specified check.
        """
        request.setHeader("content-type", "application/xml")
        return xmltodict.unparse(get_check(check_id))

    @app.route('/config/checks', methods=['GET'])
    def get_all_checks(self, request):
        """
        Return the current configuration and state of all checks.
        """
        request.setHeader("content-type", "application/xml")
        return xmltodict.unparse(get_all_checks())

    @app.route('/checks/delete/<check_id>', methods=['DELETE'])
    def delete_checks(self, request, check_id):
        """
        Delete the specified check and return 204 response code
        """
        response_code = delete_check(check_id) or 200
        request.setResponseCode(response_code)
        request.setHeader("content-type", "application/xml")
        return
